/* Copyright 2009, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Miscellaneous registers for the TS-4500.  Features:
 *   - Watchdog
 *   - GPIO in/out/direction
 *   - LEDs
 *   - RTC I2C interface
 *   - Lattice tagmem access
 *   - Random number generator 
 *   - SPI PLL phase control for SPI flash interface
 *   - Reset switch enable
 *   - Latched mode/jumper pins for bootup medium determination
 *   - scratch reg for communicating bootup device to OS from BOOTROM
 *   - Model reg (16 bit), submodel reg (4-bit), and FPGA revision (4-bit)
 *
 * In the 128-byte SBUS address space, this core appears starting at
 * address 0x60 on the TS-4500. 
 *
 * Register map:
 * base + 0x0: Model ID reg (RO) - reads 0x4500 on the TS-4500
 * base + 0x2: submodel, fpga revision, RTC and LED control - (R/W)
 *   bit 15: green LED (1 - on)
 *   bit 14: red LED (1 - on)
 *   bit 13: RTC SCL input
 *   bit 12: RTC SDA input
 *   bit 11: RTC SCL direction (1 - output)
 *   bit 10: RTC SDA direction (1 - output)
 *   bit 9: RTC SCL output
 *   bit 8: RTC SDA output
 *   bit 7-4: Board submodel (RO) - 0x0 on production TS-4500
 *   bit 3-0: FPGA revision
 * base + 0x4: 16-bits of random data changed every 1 second. (RO)
 * base + 0x6: DIO and tagmem control (RW)
 *   bit 15-12: DIO input for pins 40(MSB)-37(LSB) (RO)
 *   bit 11-8: DIO output for pins 40(MSB)-37(LSB) (RW)
 *   bit 7-4: DIO direction for pins 40(MSB)-37(LSB) (1 - output) (RW)
 *   bit 3: Lattice tagmem clock (RW)
 *   bit 2: Lattice tagmem serial-in (RW)
 *   bit 1: Lattice tagmem CSn (RW)
 *   bit 0: Lattice tagmem serial-out (RO)
 * base + 0x8: DIO input for pins 36(MSB)-21(LSB) (RO)  
 * base + 0xa: DIO output for pins 36(MSB)-21(LSB) (RW)
 * base + 0xc: DIO direction for pins 36(MSB)-21(LSB) (1 - output) (RW)
 * base + 0xe: DIO input for pins 20(MSB)-5(LSB) (RO)  
 * base + 0x10: DIO output for pins 20(MSB)-5(LSB) (RW)
 * base + 0x12: DIO direction for pins 20(MSB)-5(LSB) (1 - output) (RW)
 * base + 0x14: Watchdog feed register (write only)
 *   write value 0x0: feed watchdog for another .338s (approx)
 *   write value 0x1: feed watchdog for another 2.706s (approx)
 *   write value 0x2: feed watchdog for another 10.824 (approx)
 *   write value 0x3: disable watchdog
 * base + 0x16: SPI PLL phase, latched mode bits, scratch reg
 *   bit 15-14: reserved
 *   bit 13: Macrocontroller RS422 enable (xuart_rx[0] = dio_i[13]) (RW)
 *   bit 12: TS-8510 specific RS422 enable (xuart_rx[4] = dio_i[13]) (RW)
 *   bit 11: CAN bus enable (RW)
 *   bit 10-6: PLL phase (set by TS-BOOTROM) (RW)
 *   bit 5: mode3 latched bootstrap bit (RO)
 *   bit 4: Reset switch enable (1 - auto reboot when dio_i[9] == 0) (RW)
 *   bit 3-2: scratch reg (RW)
 *   bit 1: mode2 latched bootstrap bit (RO)
 *   bit 0: mode1 latched bootstrap bit (RO)
 * base + 0x18: DIO
 *   bit 15: reserved
 *   bit 14-10: DIO input for pins 4(MSB)-0(LSB) (RO)
 *   bit 9-5: DIO output for pins 4(MSB)-0(LSB) (RW)
 *   bit 4-0: DIO direction for pins 4(MSB)-0(LSB) (1 - output) (RW)
 * base + 0x1a: DIO
 *   bit 15-8: DIO output for pins 48(MSB)-41(LSB) (RW)
 *   bit 7-0: DIO direction for pins 48(MSB)-41(LSB) (1 - output) (RW)
 * base + 0x1c: DIO
 *   bit 15-12: reserved
 *   bit 11-0: DIO input for pins 52(MSB)-41(LSB) (RO)
 * base + 0x1e: DIO
 *   bit 15-6: reserved
 *   bit 7-4: DIO output for pints 52(MSB)-49(LSB) (RW)
 *   bit 3-0: DIO direction for pins 52(MSB)-49(LSB) (1 - output) (RW)
 *
 * * Watchdog by default comes out of reset armed for .338 seconds.  
 *   TS-BOOTROM firmwarm feeds for 10.824 and OS code has 10.824 seconds
 *   to take over.
 *
 * Revision history:
 * 0 - Initial release
 * 1 - Added another bit to SPI speed to allow lower speeds (1.2Mhz),
 *     fixed DIO pins, fixed ethernet LEDs, added CAN enable reg bit.
 * 2 - Fix CAN clock enable/disable.
 * 3 - Add TS-8510 full-duplex RS485 (RS422) enable bit to syscon.  Also
 *     add RS422 enable bit for all the other macrocontroller base boards.
 *     Update bootrom to Dec 1
 * 4 - Memwindow and muxbus support added
 *
 */

module syscon(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_we_i,
  wb_adr_i,
  wb_dat_i,
  wb_sel_i,
  wb_dat_o,
  wb_ack_o,

  dio_i,
  dio_oe_o,
  dio_o,

  rtc_sda_o,
  rtc_sda_i,
  rtc_sda_oe_o,
  rtc_scl_o,
  rtc_scl_i,
  rtc_scl_oe_o,

  internal_osc_o,
  clk_100mhz_i,
  reboot_o,
  led_grn_o,
  led_red_o,
  mode1_i,
  mode2_i,
  mode3_i,
  pllphase_o,
  en_usb_power_o,
  can_enable_o,
  can_wbaccess_i,
  en_ts8510_rs422_o,
  en_rs422_o
);

input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i;
input [31:0] wb_adr_i;
input [15:0] wb_dat_i;
input [1:0] wb_sel_i;
output [15:0] wb_dat_o;
output wb_ack_o;

input [52:0] dio_i;
output [52:0] dio_oe_o;
output [52:0] dio_o;

input rtc_sda_i, rtc_scl_i;
output rtc_sda_o, rtc_sda_oe_o, rtc_scl_o, rtc_scl_oe_o;
output led_grn_o, led_red_o;
output internal_osc_o, reboot_o;
input clk_100mhz_i, mode1_i, mode2_i, mode3_i;

output [4:0] pllphase_o;
output en_usb_power_o;

input can_wbaccess_i;
output can_enable_o;
output en_ts8510_rs422_o;
output en_rs422_o;

parameter wdog_default = 1;
parameter can_opt = 0;

localparam dio_mask = {53{1'b1}};

wire random_clk;
OSCE #(.NOM_FREQ("3.1")) rndosc (random_clk);
assign internal_osc_o = random_clk;
reg [15:0] rndcnt;
reg [1:0] wdogctl;
reg [9:0] wdogcnt;
reg feed_en, reboot, shift;
wire sed_err;
reg resetsw_en;
reg [1:0] resetsw;
wire feed_bsy, feed_rdreq;
wire [15:0] wdog_dat;
assign reboot_o = reboot | (resetsw_en && resetsw == 2'b00) /*| sed_err */;
always @(posedge random_clk or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    rndcnt <= 16'd0;
    shift <= 1'b0;
    wdogctl <= wdog_default;
    wdogcnt <= 10'd0;
    resetsw <= 2'b11;
  end else begin
    shift <= 1'b0;
    rndcnt <= rndcnt + 1'b1;
    if (rndcnt == 16'd0) begin
      shift <= 1'b1;
      resetsw[0] <= dio_i[9];
      resetsw[1] <= resetsw[0] | dio_i[9];
      wdogcnt <= wdogcnt + 1'b1;
    end
    if (feed_rdreq) begin
      wdogctl <= wdog_dat;
      wdogcnt <= 10'd0;
    end
  end
end
always @(*) begin
  case (wdogctl)
  2'd0: reboot = wdogcnt[4]; /* approx .338 seconds */
  2'd1: reboot = wdogcnt[7]; /* approx 2.706 seconds */
  2'd2: reboot = wdogcnt[9]; /* approx 10.824 seconds */
  2'd3: reboot = 1'b0;
  endcase
end

wire [15:0] random;
crc randnumgen(
  .dat_i(clk_100mhz_i),
  .clk_i(random_clk),
  .clken_i(shift),
  .rst_i(wb_rst_i),
  .shift_i(1'b0),
  .crc16_o(random)
);
/*
SEDBA sedcore (
  .SEDENABLE(1'b1),
  .SEDSTART(1'b1),
  .SEDFRCERRN(1'b1),
  .SEDERR(sed_err)
);
*/
resync_fifo watchdogfifo(
  .rst_i(wb_rst_i),
  .wrclk_i(wb_clk_i),
  .wrreq_i(feed_en),
  .data_i(wb_dat_i),
  .full_wrclk_o(feed_bsy),
  .rdclk_i(random_clk),
  .rdreq_i(feed_rdreq),
  .q_o(wdog_dat),
  .full_rdclk_o(feed_rdreq)
);

reg [6:0] count;
reg [31:0] usec;
always @(posedge clk_100mhz_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    count <= 7'd0;
    usec <= 32'd0;
  end else begin
    count <= count + 1'b1;
    if (count == 7'd99) begin
      count <= 7'd0;
      usec <= usec + 1'b1;
    end
  end
end

reg spi_csn, spi_clk, spi_si;
wire spi_so;
tagmem tagmemcore(
  .CLK(spi_clk),
  .SI(spi_si),
  .SO(spi_so),
  .CS(spi_csn)
);

reg rtc_sda, rtc_scl, rtc_sda_oe, rtc_scl_oe;
assign rtc_sda_o = rtc_sda;
assign rtc_scl_o = rtc_scl;
assign rtc_sda_oe_o = rtc_sda_oe;
assign rtc_scl_oe_o = rtc_scl_oe;
reg [52:0] dio_oe, dio;
assign dio_oe_o = dio_oe[52:0];
assign dio_o = dio[52:0];
reg [15:0] wb_dat;
assign wb_dat_o = wb_dat;
reg led_grn, led_red;
assign led_grn_o = led_grn;
assign led_red_o = led_red;
reg [1:0] scratch;
reg [4:0] pllphase;
assign en_usb_power_o = dio[7];
assign pllphase_o = pllphase;
reg can_enable;
assign can_enable_o = can_enable;
reg en_ts8510_rs422, en_rs422;
assign en_ts8510_rs422_o = en_ts8510_rs422;
assign en_rs422_o = en_rs422;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    dio_oe <= 53'd0;
    dio <= 53'h80;
    rtc_sda <= 1'b0;
    rtc_scl <= 1'b0;
    rtc_sda_oe <= 1'b0;
    rtc_scl_oe <= 1'b0;
    spi_clk <= 1'b0;
    spi_si <= 1'b0;
    spi_csn <= 1'b1;
    led_grn <= 1'b1;
    led_red <= 1'b1;
    scratch <= 2'b00;
    resetsw_en <= 1'b0;
    pllphase <= 5'h18;
    can_enable <= 1'b0;
    en_ts8510_rs422 <= 1'b0;
    en_rs422 <= 1'b0;
  end else begin
    if (wb_cyc_i && wb_stb_i && wb_we_i) case (wb_adr_i[4:0])
      5'h2: begin
        {rtc_scl_oe, rtc_sda_oe, rtc_scl, rtc_sda} <= wb_dat_i[11:8];
        {led_grn, led_red} <= wb_dat_i[15:14];
      end
      5'h6: begin
        {dio[40:37], dio_oe[40:37]} <= wb_dat_i[11:4] & {2{dio_mask[40:37]}};
        {spi_clk, spi_si, spi_csn} <= wb_dat_i[3:1];
      end
      5'ha: dio[36:21] <= wb_dat_i[15:0] & dio_mask[36:21];
      5'hc: dio_oe[36:21] <= wb_dat_i[15:0] & dio_mask[36:21];
      5'h10: dio[20:5] <= wb_dat_i[15:0] & dio_mask[20:5];
      5'h12: dio_oe[20:5] <= wb_dat_i[15:0] & dio_mask[20:5];
      5'h16: begin
        en_rs422 <= wb_dat_i[13];
        en_ts8510_rs422 <= wb_dat_i[12];
        if (can_opt) can_enable <= wb_dat_i[11];
        pllphase <= wb_dat_i[10:6];
        scratch <= wb_dat_i[3:2];
        resetsw_en <= wb_dat_i[4];
      end
      5'h18: {dio[4:0], dio_oe[4:0]} <= wb_dat_i[9:0] & {2{dio_mask[4:0]}};
      5'h1a: begin
        dio[48:41] <= wb_dat_i[15:8] & dio_mask[48:41];
        dio_oe[48:41] <= wb_dat_i[7:0] & dio_mask[48:41];
      end
      5'h1e: 
        {dio[52:49], dio_oe[52:49]} <= wb_dat_i[7:0] & {2{dio_mask[52:49]}};
    endcase
    if (can_wbaccess_i && can_opt) can_enable <= 1'b1;
  end
end

reg wb_ack;
assign wb_ack_o = wb_ack;
wire [52:0] dio_in;
assign dio_in[52:0] = dio_i[52:0] & dio_mask[52:0];
always @(*) begin
  wb_ack = wb_cyc_i && wb_stb_i;
  wb_dat = 16'hxxxx;
  feed_en = 1'b0;

  case (wb_adr_i[4:0]) 
  5'h0: wb_dat = 16'h4500; /* Model ID */ 
  5'h2: wb_dat = {led_grn, led_red, rtc_scl_i, rtc_sda_i, rtc_scl_oe,
    rtc_sda_oe, rtc_scl, rtc_sda, 4'h0, 4'h4}; /* submodel, revision */
  5'h4: wb_dat = random[15:0];
  5'h6: wb_dat = {dio_in[40:37], dio[40:37], dio_oe[40:37], spi_clk, spi_si, spi_csn, spi_so};
  5'h8: wb_dat = dio_in[36:21];
  5'ha: wb_dat = dio[36:21];
  5'hc: wb_dat = dio_oe[36:21];
  5'he: wb_dat = dio_in[20:5];
  5'h10: wb_dat = dio[20:5];
  5'h12: wb_dat = dio_oe[20:5];
  5'h14: begin
    feed_en = wb_cyc_i & wb_stb_i & wb_we_i & !feed_bsy;
    wb_ack = wb_cyc_i & wb_stb_i & ((wb_we_i & !feed_bsy) | !wb_we_i);
  end
  5'h16: wb_dat = {{3{1'bx}}, en_rs422, en_ts8510_rs422, can_enable, pllphase,
    mode3_i, resetsw_en, scratch, mode2_i, mode1_i};
  5'h18: wb_dat = {1'bx, dio_in[4:0], dio[4:0], dio_oe[4:0]};
  5'h1a: wb_dat = {dio[48:41], dio_oe[48:41]};
  5'h1c: wb_dat = {4'hx, dio_in[52:41]};
  5'h1e: wb_dat = {{8{1'bx}}, dio[52:49], dio_oe[52:49]};
  endcase

end


endmodule

