/* Copyright 2007-2009, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* NAND flash controller with hardware Hamming-code ECC generator.
 * ECC generator has 8 entry FIFO to allow for streaming of an entire
 * 2 kbyte page while saving and reseting the ECC syndrome every 256
 * bytes.
 *
 * Register map:
 * base + 0x0: NAND control and ECC register
 *   bit 0: ALE, address latch enable bit (1 - asserted)
 *   bit 1: CLE, command latch enable bit (1 - asserted)
 *   bit 2: CS, chip select (1 - asserted)
 *   bit 3: ECC fifo enable
 *   bit 4: FIFO last or FIFO empty (RO)
 *   bit 5: NAND chip busy, (1 - busy) (RO)
 *   bit 7-6: reserved
 *   bit 31-8: Hamming code ECC bits or BOOTROM scrub data (if ecc_opt == 0)
 * base + 0x4: NAND data register
 * base + 0x8: RTC address register
 * base + 0xc: RTC data register
 * base + 0x10: CAN address register
 * base + 0x14: CAN data register
 *
 * * When FIFO is enabled, bits 31-8 of register 0x0 return enqueued
 *   ECC bits until a read with bit 4 set.  When bit 4 is set, the 
 *   read ECC bits are valid, but the FIFO is now empty and further
 *   reads will be UNDEFINED.
 * 
 * * If the FIFO is not enabled, bits 31-8 of register 0x0 will return
 *   current ECC state.
 *
 * * Reads or writes to bits 31-24 of register 0x0 will advance the FIFO
 *   (if FIFO is enabled) and reset ECC state.
 * 
 * * A 32-bit read at the data register will cause 4 NAND RD# cycles
 *   and return the 4 bytes as a little endian (LSB first) word.  A
 *   16-bit read will return 2 cycles, and a 8-bit 1.  8-bit and
 *   16-bit reads must only be to base + 0x4 address, reads/writes
 *   to base + 5, 6, 7 are UNDEFINED. 
 *
 * * RTC and CAN registers are 8-bit only.
 */
/* Timing requirements with wishbone bus @100Mhz. 30ns strobe, 20ns idle
 * (50ns cycles), input sample point @40ns, output updated @0ns. PCB
 * trace skew helps FPGA hold, hurts FPGA setup and does nothing for NAND
 * setup/hold assuming equal length clk and data paths. Bus capacitance 
 * above 50pF hurts FPGA setup, helps FPGA hold and vice versa for
 * bus capacitance below 50pF.
 * 
 * Outputs (NAND flash has 15ns tSU and 5ns hold requirement)
 * * max data tCO - max write tCO < 15ns (NAND setup margin) 
 * * min data tCO - min write tCO + 20ns > 5ns (NAND hold margin)
 *
 * Inputs (NAND flash has 25ns tCO and guarantees 15ns hold time)
 * * max read tCO + 25ns + data tSU < 40ns (FPGA setup margin)
 * * data tHOLD - min read tCO + 10ns < 15ns (FPGA hold margin)
 *
 * TS-7390 NAND margins as of 8/19/2008 (50pF bus loads)
 * NAND setup margin: 16.344ns
 * NAND hold margin: 10.941ns
 * FPGA hold margin: 7.501ns
 * FPGA setup margin: 3.109ns
 */

module nandflash(
  wb_clk_i,
  wb_rst_i,

  wb1_cyc_i,
  wb1_cti_i,
  wb1_bte_i,
  wb1_stb_i,
  wb1_we_i,
  wb1_adr_i,
  wb1_dat_i,
  wb1_sel_i,
  wb1_dat_o,
  wb1_ack_o,

  wb2_cyc_i,
  wb2_stb_i,
  wb2_cti_i,
  wb2_bte_i,
  wb2_we_i,
  wb2_adr_i,
  wb2_dat_i,
  wb2_sel_i,
  wb2_dat_o,
  wb2_ack_o,

  wb3_cyc_i,
  wb3_stb_i,
  wb3_cti_i,
  wb3_bte_i,
  wb3_we_i,
  wb3_adr_i,
  wb3_dat_i,
  wb3_sel_i,
  wb3_dat_o,
  wb3_ack_o,

  wb4_cyc_i,
  wb4_stb_i,
  wb4_cti_i,
  wb4_bte_i,
  wb4_we_i,
  wb4_adr_i,
  wb4_dat_i,
  wb4_sel_i,
  wb4_dat_o,
  wb4_ack_o,


  nand_cle_o,
  nand_ale_o,
  nand_cs_o,
  nand_rd_o,
  nand_wr_o,
  nand_busy_i,
  nand_dat_i,
  nand_dat_o,
  nand_dat_oe_o,
  rtc_cs_o,
  can_cs_o
);

input wb_clk_i, wb_rst_i;

input wb1_cyc_i, wb1_stb_i, wb1_we_i;
input [3:0] wb1_sel_i;
input [31:0] wb1_adr_i, wb1_dat_i;
output [31:0] wb1_dat_o;
output wb1_ack_o;
input [2:0] wb1_cti_i;
input [1:0] wb1_bte_i;

input wb2_cyc_i, wb2_stb_i, wb2_we_i;
input [3:0] wb2_sel_i;
input [31:0] wb2_adr_i, wb2_dat_i;
output [31:0] wb2_dat_o;
output wb2_ack_o;
input [2:0] wb2_cti_i;
input [1:0] wb2_bte_i;

input wb3_cyc_i, wb3_stb_i, wb3_we_i;
input [3:0] wb3_sel_i;
input [31:0] wb3_adr_i, wb3_dat_i;
output [31:0] wb3_dat_o;
output wb3_ack_o;
input [2:0] wb3_cti_i;
input [1:0] wb3_bte_i;

input wb4_cyc_i, wb4_stb_i, wb4_we_i;
input [3:0] wb4_sel_i;
input [31:0] wb4_adr_i, wb4_dat_i;
output [31:0] wb4_dat_o;
output wb4_ack_o;
input [2:0] wb4_cti_i;
input [1:0] wb4_bte_i;

output nand_cle_o, nand_ale_o, nand_cs_o, nand_rd_o, nand_wr_o;
output [7:0] nand_dat_o;
output nand_dat_oe_o;
input [7:0] nand_dat_i;
input nand_busy_i;
output rtc_cs_o;
output can_cs_o;

endmodule
